/*
* MODBUS.c
*
* Created: 11.04.2022 13:19:16
*  Author: tomas.viktora
*/
#include "MODBUS.h"
#include "std.h"


uint make_int(uchar HB, uchar LB)
{
	uint16_t temp = 0;

	temp = HB;
	temp <<= 8;
	temp |= LB;

	return temp;
}

void flush_rx_data(void)
{
	signed char i = (32 - 1);

	while(i > -1)
	{
		rx_data[i--] = 0x00;
	};

	cnt_rx = 0x00;
}


void flush_tx_data(void)
{
	signed char i = (32 - 1);

	while(i > -1)
	{
		tx_data[i--] = 0x00;
	};

	cnt_tx = 0x00;
}

uint16_t check_crc(uint16_t data_length, uchar *data)
{
	uint16_t crc_word = 0xFFFF;
	
	unsigned int i = 0;
	char bit = 0;

	for( i = 0; i < data_length; i++ )
	{
		crc_word ^= data[i];

		for( bit = 0; bit < 8; bit++ )
		{
			if( crc_word & 0x0001 )
			{
				crc_word >>= 1;
				crc_word ^= 0xA001;
			}
			else
			{
				crc_word >>= 1;
			}
		}
	}

	return crc_word;
}

void get_hb_lb(uint value, uchar *HB, uchar *LB)
{
	*LB = (value & 0x00FF);
	*HB = ((value & 0xFF00) >> 0x08);
}

void recieve_task(void)
{
	static uchar data_array[0x18] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	uint16_t start_register = 0x0000;
	uint16_t no_of_register = 0x0000;
	uint16_t state_digital_out = 0x0000;
	uint16_t state_analog_out = 0x0000;
	uint16_t crc = 0x0000;

	uint error_label = 0;
	

	if (rx_data[ID_BYTE] == SLAVE_ID)
	{
		switch(rx_data[FUNCTION_BYTE])
		{
			case READ_DO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);

				if((start_register >= 0) && (start_register <= 31))
				{
					if((no_of_register <= 32) && ((start_register + no_of_register -1) <= 31))
					{
						if(crc == check_crc(6, rx_data))
						{
							data_array[0] = 0x01;

							if (no_of_register > 8)
							{
								data_array[0] = 0x02;
							}

							if (no_of_register > 16)
							{
								data_array[0] = 0x03;
							}

							if (no_of_register > 24)
							{
								data_array[0] = 0x04;
							}
							
							uint no_of_bytes = (uint)data_array[0];
							data_array[1] = digital_out[0];
							data_array[2] = digital_out[1];
							data_array[3] = digital_out[2];
							data_array[4] = digital_out[3];

							send_task(READ_DO, no_of_bytes + 1, data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;


				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(READ_DO | 0x80, 1, data_array);
				}
				
				break;
			}

			case READ_DI:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);

				if((start_register >= 0) && (start_register <= 31))
				{
					if((no_of_register <= 32) && ((start_register + no_of_register -1) <= 31))
					{
						if(crc == check_crc(6, rx_data))
						{
							data_array[0] = 0x01;

							if (no_of_register > 8)
							{
								data_array[0] = 0x02;
							}

							if (no_of_register > 16)
							{
								data_array[0] = 0x03;
							}

							if (no_of_register > 24)
							{
								data_array[0] = 0x04;
							}
							
							uint no_of_bytes = (uint)data_array[0];
							data_array[1] = digital_in[0];
							data_array[2] = digital_in[1];
							data_array[3] = digital_in[2];
							data_array[4] = digital_in[3];

							send_task(READ_DI, no_of_bytes + 1, data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(READ_DI | 0x80, 1, data_array);
				}
				
				break;
			}

			case READ_AO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);

				uchar hb = 0x00;
				uchar lb = 0x00;

				if((start_register >= 0) && (start_register <= 7))
				{
					if((no_of_register <= 8) && ((start_register + no_of_register -1) <= 7))
					{
						if(crc == check_crc(6, rx_data))
						{
							data_array[0] = (2 * no_of_register);

							int j = 1;

							for (int i = 1; i < (2 * no_of_register) + 1; i++)
							{
								hb = 0x00;
								lb = 0x00;
																
								get_hb_lb(analog_out[start_register + i - 1], &hb, &lb);
								data_array[j] = hb;			
								j++;									
								data_array[j] = lb;
								j++;
							}
							
							send_task(READ_AO, ((2 * no_of_register) + 1), data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(READ_AO | 0x80, 1, data_array);
				}
				break;
			}

			case READ_AI:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);

				uchar hb = 0x00;
				uchar lb = 0x00;

				if((start_register >= 0) && (start_register <= 7))
				{
					if((no_of_register <= 8) && ((start_register + no_of_register -1) <= 7))
					{
						if(crc == check_crc(6, rx_data))
						{
							data_array[0] = (2 * no_of_register);

							int j = 1;

							for (int i = 1; i < no_of_register + 1; i++)
							{
								hb = 0x00;
								lb = 0x00;
								
								get_hb_lb(analog_in[start_register + i - 1], &hb, &lb);
								data_array[j] = hb;
								j++;
								data_array[j] = lb;
								j++;
							}
							
							send_task(READ_AI, ((2 * no_of_register) + 1), data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(READ_AI | 0x80, 1, data_array);
				}
				break;
			}

			case WRITE_SINGLE_DO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				state_digital_out = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);
				uchar temp_state = 0x01;

				if((start_register >= 0) && (start_register <= 31))
				{

					if((state_digital_out == DO_SET_ON) ||(state_digital_out == DO_SET_OFF))
					{
						
						if(crc == check_crc(6, rx_data))
						{
							
							if(state_digital_out == DO_SET_ON)
							{
								
								if(start_register > 0x07)
								{
									temp_state = start_register - 8;
									digital_out[1] |= temp_state;
								}
								if(start_register > 0x0F)
								{
									temp_state <<= start_register - 16;
									digital_out[2] |= temp_state;
								}
								if(start_register > 0x17)
								{
									temp_state <<= start_register - 24;
									digital_out[3] |= temp_state;
								}
								else
								{
									temp_state <<= start_register;
									digital_out[0] |= temp_state;
								}

							}
							else
							{
								if(start_register > 0x07)
								{
									temp_state <<= start_register - 8;
									digital_out[1] &= (~temp_state);
								}
								if(start_register > 0x0F)
								{
									temp_state <<= start_register - 16;
									digital_out[2] &= (~temp_state);
								}
								if(start_register > 0x17)
								{
									temp_state <<= start_register - 24;
									digital_out[3] &= (~temp_state);
								}
								else
								{
									temp_state <<= start_register;
									digital_out[0]&= (~temp_state);
								}
							}

							data_array[0] = rx_data[2];
							data_array[1] = rx_data[3];
							data_array[2] = rx_data[4];
							data_array[3] = rx_data[5];

							is_change_out = 1;

							send_task(WRITE_SINGLE_DO, 4, data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(WRITE_SINGLE_DO | 0x80, 1, data_array);
				}

				break;
			}

			case WRITE_SINGLE_AO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				state_analog_out = make_int(rx_data[4],rx_data[5]);
				crc = make_int(rx_data[7],rx_data[6]);			

				if((start_register >= 0) && (start_register <= 7))
				{
					if(crc == check_crc(6, rx_data))
					{
						analog_out[start_register]= state_analog_out;
						
						data_array[0] = rx_data[2];
						data_array[1] = rx_data[3];
						data_array[2] = rx_data[4];
						data_array[3] = rx_data[5];

						send_task(WRITE_SINGLE_AO, 4, data_array);
					}
					else
						error_label = -1;
				}
				else
					error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(WRITE_SINGLE_AO | 0x80, 1, data_array);
				}

				break;
			}

			case WRITE_MULTIPLE_DO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				uint no_of_bytes = rx_data[6];
				crc = make_int(rx_data[8 + no_of_bytes],rx_data[7 + no_of_bytes]);

				if((start_register >= 0) && (start_register <= 31))
				{
					if((no_of_register <= 32) && ((start_register + no_of_register -1) <= 31))
					{
						if(crc == check_crc(7 + no_of_bytes, rx_data))
						{
							for(int i = 0; i < no_of_register; i++)
							{
								digital_out[(start_register + i) / 8] &= ~(0b00000001 << ((start_register + i) % 8)); // Vynulovan bitu
								digital_out[(start_register + i) / 8] |= (((rx_data[7 + (i / 8)] >> (i % 8)) & 0b00000001) << ((start_register + i) % 8)); // Zapsn 0 nobe 1 do na sprvn msto
							}
							
							data_array[0] = rx_data[2];
							data_array[1] = rx_data[3];
							data_array[2] = rx_data[4];
							data_array[3] = rx_data[5];
							
							send_task(WRITE_MULTIPLE_DO, 4, data_array);

							is_change_out = 1;
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(WRITE_MULTIPLE_DO | 0x80, 1, data_array);
				}				
				break;
			}

			case WRITE_MULTIPLE_AO:
			{
				start_register = make_int(rx_data[2],rx_data[3]);
				no_of_register = make_int(rx_data[4],rx_data[5]);
				uint no_of_bytes = rx_data[6];
				crc = make_int(rx_data[8 + no_of_bytes],rx_data[7 + no_of_bytes]);
				
				if((start_register >= 0) && (start_register <= 7))
				{
					if((no_of_register <= 8) && ((start_register + no_of_register -1) <= 7))
					{
						if(crc == check_crc(7 + no_of_bytes, rx_data))
						{
							for(int i = 0; i < no_of_register;i++)							
								analog_out[i + start_register] = make_int(rx_data[7 + i], rx_data[8 + i]);													
						
							data_array[0] = rx_data[2];
							data_array[1] = rx_data[3];
							data_array[2] = rx_data[4];
							data_array[3] = rx_data[5];							
							send_task(WRITE_MULTIPLE_AO, 4, data_array);
						}
						else
						error_label = -1;
					}
					else
					error_label = -1;
				}
				else
				error_label = -1;

				if(error_label == -1)
				{
					data_array[0] = 0x02;
					send_task(WRITE_MULTIPLE_AO | 0x80, 1, data_array);
				}
				break;
			}

			default:
			{
				data_array[0] = 0x01;
				send_task(rx_data[FUNCTION_BYTE] | 0x80, 1, data_array);
				break;
			}
		}
	}

	flush_rx_data();
}

void send_task(uchar function_code, uchar data_length, uchar *data)
{
	uchar crc_hb = 0x00;
	uchar crc_lb = 0x00;
	uint lenght_sended_data = 0;

	flush_tx_data();

	tx_data[ID_BYTE] = SLAVE_ID;
	tx_data[FUNCTION_BYTE] = function_code;

	for(int i = 0; i < data_length; i++)
	{
		tx_data[2+i] = data[i];
	}

	get_hb_lb(check_crc((data_length + NO_OF_STATIC_BYTES),tx_data), &crc_hb, &crc_lb);
	
	tx_data[DATA_START_BYTE + data_length] = crc_lb;
	tx_data[DATA_START_BYTE + data_length + 1] = crc_hb;

	lenght_sended_data = NO_OF_STATIC_BYTES + data_length + NO_OF_CRC_BYTE + 2;

	// Sending response

	write(TEN, 1);
	UCSR0B &= ~(1<<RXEN0);

	for (int i = 0; i < lenght_sended_data; i++)
	{
		while(!(UCSR0A &  (1<<UDRE0)));
		UDR0 = tx_data[i];
		while(!(UCSR0A &  (1<<TXC0)));
	}

	write(TEN, 0);
	UCSR0B |= 1<<RXEN0;

}