/* _____________________________________________________________________________
 *
 *	AVR BASIC IO PORT CONTROLS
 * _____________________________________________________________________________
 *
 *	Ing. Lukas PIVONKA, www.lucass.cz
 *
 *  12/5/2009
 *		- new defines strobe and strobe_us
 *		- norm_bool define renamed to nbool
 *
 */

#ifndef _IOCTRL_H_
#define _IOCTRL_H_

// _____DECLARATIONS____________________________________________________________

// joins two parameters
#define concat(a, b) a ## b
// normalize a boolean, any value that corresponds to logical one is converted to number "1"
#define nbool(x) ((x)?1:0)
// set pin as output
#define setout_nodef(port, pin_no) (concat(DDR, port) |= _BV(pin_no))
#define setout(define) setout_nodef(define)
// set pin as input
#define setin_nodef(port, pin_no) (concat(DDR, port) &= ~_BV(pin_no))
#define setin(define) setin_nodef(define)
// read port pin no value
#define read_port_nodef(port, pin_no) (concat(PORT, port) & _BV(pin_no))
#define read_port(define) read_port_nodef(define)
// read pin no value
#define read_nodef(port, pin_no) (concat(PIN, port) & _BV(pin_no))
#define read(define) read_nodef(define)
// write port pin no value
#define write_nodef(port, pin_no, value) value?(concat(PORT, port) |= _BV(pin_no)):(concat(PORT, port) &= ~_BV(pin_no))
#define write(define, value) write_nodef(define, value)
// invert port pin no value
#define invert_nodef(port, pin_no) concat(PORT, port) ^= _BV(pin_no)
#define invert(define) invert_nodef(define)
// strobe pulse for three clocks (instruction time), level of the pulse depends on previous port value (zero = positive pulse)
#define strobe_nodef(port, pin_no) concat(PORT, port) ^= _BV(pin_no); concat(PORT, port) ^= _BV(pin_no)
#define strobe(define) strobe_nodef(define)
// strobe pulse for a number of microseconds (util/delay.h must be included and also F_CLK must be set for using this define)
#define strobe_us_nodef(port, pin_no, delay) concat(PORT, port) ^= _BV(pin_no); _delay_us(delay); concat(PORT, port) ^= _BV(pin_no)
#define strobe_us(define, delay) strobe_us_nodef(define, delay)


// _____________________________________________________________________________

#endif
